/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.enterprise.impl;

import java.io.Serializable;
import java.util.logging.Level;
import org.jbpm.log.Log;

import javax.ejb.CreateException;
import javax.ejb.EJBException;
import javax.ejb.MessageDrivenBean;
import javax.ejb.MessageDrivenContext;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageListener;
import javax.jms.ObjectMessage;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.jbpm.PvmException;
import org.jbpm.client.Command;
import org.jbpm.enterprise.LocalCommandExecutor;
import org.jbpm.enterprise.LocalCommandExecutorHome;
import org.jbpm.jobexecutor.ExecuteJobCommand;

/**
 * This message-driven bean listens for ObjectMessages that contain a
 * {@link Command} object. The commands that are received are executed by a
 * LocalCommandService bean (using the local interface). The local command
 * service can be configured in the deployment descriptor. The local reference
 * used is java:comp/env/ejb and then LocalCommandServiceHome.
 * 
 * @author Jim Rigsbee
 * @author Tom Baeyens
 */
public class CommandReceiverMDB implements MessageDrivenBean, MessageListener {

  private MessageDrivenContext messageDrivenContext;
  private LocalCommandExecutor commandExecutor;

  private static final long serialVersionUID = 1L;

  private static final Log log = Log.getLog(CommandReceiverMDB.class.getName());

  public void onMessage(Message message) {
    try {
      Command<?> command = extractCommand(message);
      commandExecutor.execute(command);
    }
    catch (JMSException e) {
      log.error("couldn't process message " + message, e);
      messageDrivenContext.setRollbackOnly();
      throw new PvmException("couldn't process message " + message, e);
    }
  }

  protected Command<?> extractCommand(Message message) throws JMSException {
    Command<?> command = null;

    // messages can contain serialized commands
    if (message instanceof ObjectMessage) {
      log.trace("deserializing command from jms message...");
      ObjectMessage objectMessage = (ObjectMessage) message;
      Serializable object = objectMessage.getObject();
      if (object instanceof Command) {
        command = (Command<?>) object;
      }
      else {
        /*
         * TODO verify that such a message gets redirected to the dead letter
         * queue or something similar without retrying.
         */
        log.info("WARNING: ignoring object message cause it isn't a command: " + object);
      }
    }
    else {
      // messages can also contain a jobId property
      // if jobId is available
      if (message.getObjectProperty("jobId") != null) {
        log.trace("getting job id from jms message...");
        long jobId = message.getLongProperty("jobId");
        log.trace("retrieved jobId '" + jobId + "' via jms message");

        // create a command with the given jobId
        command = new ExecuteJobCommand(jobId);
      }
      else {
        log.info("WARNING: JobListenerBean is ignoring message '"
            + message
            + "' that doesn't have jobId property");
      }
    }
    return command;
  }

  public void setMessageDrivenContext(MessageDrivenContext messageDrivenContext) {
    this.messageDrivenContext = messageDrivenContext;
  }

  public void ejbCreate() {
    try {
      Context context = new InitialContext();
      LocalCommandExecutorHome commandExecutorHome = (LocalCommandExecutorHome) context.lookup("java:comp/env/ejb/LocalCommandExecutor");
      context.close();

      commandExecutor = commandExecutorHome.create();
    }
    catch (NamingException e) {
      throw new EJBException("error retrieving command executor home", e);
    }
    catch (CreateException e) {
      throw new EJBException("error creating command executor", e);
    }
  }

  public void ejbRemove() {
    commandExecutor = null;
    messageDrivenContext = null;
  }
}
