/**
 * Copyright (C) 2007  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.jbpm.env.session;

import java.util.ArrayList;
import java.util.Collection;
import org.jbpm.log.Log;

import org.jbpm.PvmException;
import org.jbpm.env.Environment;
import org.jbpm.env.EnvironmentFactory;
import org.jbpm.jobexecutor.JobExecutor;
import org.jbpm.jobexecutor.JobSession;
import org.jbpm.pvm.Activity;
import org.jbpm.pvm.Execution;
import org.jbpm.pvm.impl.ObjectReference;
import org.jbpm.tx.StandardTransaction;
import org.jbpm.tx.Transaction;
import org.jbpm.util.Listener;

/**
 * Timers created with this service are committed at the end of the transaction,
 * so their execution will be late if the delay is shorter than the transaction.
 * In that case, they will be executed at the end of the transaction.
 * 
 * @author Pascal Verdage
 */
public class StandardTimerSession implements TimerSession {
	private static final Log log = Log.getLog(TimerSession.class.getName());

	//this method is called by the binding
	public void start() {
		final EnvironmentFactory environmentFactory =
			Environment.getCurrent().getEnvironmentFactory();
		Listener jobNotificator = new Listener() {
			public void event(Object source, String eventName, Object info) {
				if ( (Transaction.EVENT_AFTERCOMPLETION.equals(eventName))
				&& (StandardTransaction.STATE_COMMITTED.equals(info)) ) {
					log.trace("awaking JobExecutor");
					environmentFactory.openEnvironment();
					try {
						Environment.getCurrent().get(JobExecutor.class).jobWasAdded();
					} finally {
						Environment.getCurrent().close();
					}
				}
			}
		};
		Environment.getCurrent().get(Transaction.class)
		.addListener(jobNotificator);
	}

	protected static void validScheduling(Timer timer) {
		if (timer == null)
			throw new IllegalArgumentException("null timer scheduled");
		if (timer.getActivityReference() == null)
			throw new IllegalArgumentException("null activity scheduled");
		if (timer.getDueDate() == null)
			throw new IllegalArgumentException("timer scheduled at null date");
		if (timer.getDueDate().getTime() < 0)
			throw new IllegalArgumentException("timer scheduled with a negative date");
		if (timer.getDueDate().getTime() == Long.MAX_VALUE)
			throw new IllegalArgumentException("timer has invalid dueDate");
	}

	public void schedule(Timer timer) {
		validScheduling(timer);
		log.debug("scheduling " + timer.getActivityReference() +
				" for " + timer.getDueDate() +
				" with repeat every " + timer.getRepeat());
		log.trace("Timer will be scheduled at the end of the transaction");

		// then persist the job
		Environment.getCurrent().get(JobSession.class).save(timer);
	}

	public void cancel(Timer timer) {
		log.debug("canceling " + timer);
		if (timer != null) {
			log.trace("Timer will be cancelled at the end of the transaction");
			Environment.getCurrent().get(JobSession.class).delete(timer);
		} else {
			throw new PvmException("Timer is null");
		}
	}

	private Timer getNextTask() {
		return Environment.getCurrent()
		.get(JobSession.class).findNextTimer();
	}

	public ObjectReference<Activity> getNextScheduledActivity() {
		ObjectReference<Activity> result = null;
		Timer timer = getNextTask();
		if (timer != null)
			result = timer.getActivityReference();
		return result;
	}

	public long getNextScheduledActivityTime() {
		long result = -1;
		Timer timer = getNextTask();
		if (timer != null && timer.getDueDate() != null)
			result =
				timer.getDueDate().getTime() -
				System.currentTimeMillis();
		return result;
	}

	public Collection<ObjectReference<Activity>> getScheduledActivities() {
		Collection<ObjectReference<Activity>> result =
			new ArrayList<ObjectReference<Activity>>();
		Collection<Timer> timers =
			Environment.getCurrent().get(JobSession.class)
			.findAllTimers();
		for (Timer timer : timers)
			result.add(timer.getActivityReference());
		return result;
	}

	public ObjectReference<Activity> getScheduledActivity(long id) {
		ObjectReference<Activity> result = null;
		Job<?> job =
			Environment.getCurrent().get(JobSession.class).get(id);
		Timer timer = null;
		if (job instanceof Timer) {
			timer = (Timer) job;
		}

		if (timer != null)
			result = timer.getActivityReference();

		return result;
	}

	public Collection<Timer> findTimers(Execution execution) {
		Collection<Timer> result =
			new ArrayList<Timer>();
		Collection<Timer> timers =
			Environment.getCurrent().get(JobSession.class)
			.findTimers(execution);
		for (Timer timer : timers) {
			result.add(timer);
		}
		return result;
	}

}
