/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.jobexecutor;

import java.util.Date;
import java.util.logging.Level;
import org.jbpm.log.Log;

import org.jbpm.PvmException;
import org.jbpm.client.Command;
import org.jbpm.client.CommandService;
import org.jbpm.env.Environment;
import org.jbpm.env.session.Job;
import org.jbpm.tx.Transaction;


/**
 * @author Tom Baeyens
 */
public class ExecuteJobCommand implements Command<Object> {
  private static final long serialVersionUID = 1L;

  private static final Log log = Log.getLog(ExecuteJobCommand.class.getName());
  
  protected long jobDbid;

  public ExecuteJobCommand(long jobDbid) {
    this.jobDbid = jobDbid;
  }

  public Object execute(Environment environment) throws Exception {
	Job<?> job = environment.get(JobSession.class).get(jobDbid);

	// in case of exclusive jobs, the job might have been deleted
	// before we execute it (they are in a list)
	if (job != null) {
	    try {
	      log.debug("executing job "+job+"...");
	      job.execute(environment);
	
	      // if this job is locked too long, it could be unlocked by the lockmonitor and 
	      // executed by another thread.
	      Date lockExpirationDate = job.getLockExpirationTime();
	      //can be null if it was rescheduled
	      if (lockExpirationDate != null) {
		      long lockExpiration = lockExpirationDate.getTime();
		      long currentTime = System.currentTimeMillis();
		      if (currentTime>lockExpiration) {
		        throw new PvmException("job took too long: lock expired "+(currentTime-lockExpiration)+"ms ago");
		      }
	      }
	    } catch (Throwable exception) {
	      log.error("exception while executing '"+job+"'", exception);
	      handleJobExecutionException(environment, job, exception);
	    }
	
	    log.debug("executed job "+job);
	} else {
		log.debug("job " + jobDbid + "no longer exists");
	}
    return null;
  }

  /** This transaction will be marked for rollback.  A command will be associated with the 
   * Transaction.EVENT_AFTERCOMPLETION (after the job locks of the current transaction are 
   * released).  Then the command will update the Job with the exception details in a separate 
   * transaction. */
  protected void handleJobExecutionException(Environment environment, Job<?> job, Throwable exception) {
    Transaction transaction = environment.getTransaction();
    transaction.setRollbackOnly();

    CommandService commandService = (CommandService) environment.get(CommandService.class);
    JobExceptionHandler jobExceptionHandler = new JobExceptionHandler(job.getDbid(), exception, commandService);
    transaction.addListener(jobExceptionHandler, Transaction.EVENT_AFTERCOMPLETION);
  }
}
