/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.pvm;

import java.io.Serializable;

/** piece of Java code that is to be included in a process as the runtime 
 * node behaviour or as a hidden listener to process events.
 * 
 * @author Tom Baeyens
 */
public interface Activity extends Serializable {
  
  /** piece of Java code that is to be included in a process as node behaviour 
   * or as a hidden listener to process events.
   * 
   * <p>Activities can be used to implement the behaviour of nodes, in 
   * which case this behaviour is associated to a graphical node in the diagram,
   * or they can be added as events, in that case, they are being hidden from 
   * the diagram.
   * </p>
   * 
   * <p>If an activity is the node behaviour, then it can control the propagation 
   * of execution.  Node behaviours can be external activities.  That means their 
   * runtime behaviour is a wait state.  In that case, {@link ExternalActivity} 
   * should be implemented to also handle the external triggers. 
   * </p>
   * 
   * <p>If an activity is being used as a listener to process events, the propagation 
   * of the execution can not be controlled in an activity.  So following methods 
   * are not to be called during {@link Event}s.
   * <ul>
   *   <li>{@link Execution#cancel()}</li>
   *   <li>{@link Execution#end()}</li>
   *   <li>{@link Execution#end(String)}</li>
   *   <li>{@link Execution#end(String, boolean)}</li>
   *   <li>{@link Execution#execute(Node)}</li>
   *   <li>{@link Execution#move(Node)}</li>
   *   <li>{@link Execution#resume()}</li>
   *   <li>{@link Execution#signal()}</li>
   *   <li>{@link Execution#signal(java.util.Map)}</li>
   *   <li>{@link Execution#signal(String)}</li>
   *   <li>{@link Execution#signal(String, java.util.Map)}</li>
   *   <li>{@link Execution#suspend()}</li>
   *   <li>{@link Execution#takeDefaultTransition()}</li>
   *   <li>{@link Execution#take(String)}</li>
   *   <li>{@link Execution#take(Transition)}</li>
   *   <li>{@link Execution#waitForSignal()}</li>
   * </ul>
   * will result in a PvmException. 
   * </p>
   *   
   * <p>If an exception is thrown that is not handled by an 
   * {@link ExceptionHandler}, the exception will be wrapped and propagated 
   * to the original client that performed an operation on the execution.
   * </p>
   */
  void execute(Execution execution) throws Exception;
}
