/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.pvm;

import java.io.Serializable;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jbpm.PvmException;
import org.jbpm.env.Environment;
import org.jbpm.pvm.log.ProcessLog;


/**
 * a runtime path of execution.
 * 
 * <h3 id="state">State of an execution</h3>
 * 
 * <p>The state of an execution is either active or locked.  An active execution is either 
 * executing or waiting for an external trigger.  If an execution is not in {@link #STATE_ACTIVE},
 * then it is locked.  A locked execution is read only.  
 * </p>
 * 
 * <img src="execution.states.png" />
 * 
 * <p>When a new execution is created, it is in {@link #STATE_ACTIVE}.  To change 
 * the state to a locked state, use {@link #lock(String)}.  
 * {@link #STATE_ACTIVE Some STATE_* constants} are provided that represent the 
 * most commonly used locked states.  But the state '...' in the picture indicates 
 * that any string can be provided as the state in the lock method.
 * </p>
 * 
 * <p>If an execution is locked, methods that change the execution will throw
 * a {@link PvmException} and the message will reference the actual locking state.
 * {@link #fire(String, ObservableElement) Firing events}, 
 * {@link #setVariable(String, Object) updating variables}, 
 * {@link #setPriority(int) updating priority} and 
 * {@link #createComment(String) adding comments} are not considered to change an 
 * execution.  Also {@link #createExecution(String) creation} and 
 * {@link #removeExecution(Execution) removal} of child executions are unchecked, 
 * which means that those methods can be invoked by external API clients and 
 * node behaviour methods, even while the execution is in a locked state. 
 * </p>    
 * 
 * <p>Make sure that comparisons between {@link #getState()} and the 
 * {@link #STATE_ACTIVE STATE_* constants} are  
 * done with .equals and not with '==' because if executions are 
 * loaded from persistent storage, a new string is created instead 
 * of the constants.
 * </p>
 * 
 * <h3>Comments</h3>
 * 
 * @author Tom Baeyens
 */
public interface Execution extends Serializable {

  /** either executing or in a wait state waiting for a signal.
   * This is the normal state of an execution and the initial state 
   * when creating a new execution. Make sure that comparisons are 
   * done with .equals and not with '==' because if executions are 
   * loaded from persistent storage, a new string is created instead 
   * of the constants. */
  final String STATE_ACTIVE = "active";
  
  /** parents with concurrent child executions are inactive.  
   * When an execution has concurrent child executions, it implies that 
   * this execution can't be active.  For example, at a fork, the parent
   * execution can wait inactively in the fork being till all the
   * child executions are joined.  Only leaves of the
   * execution tree can be active. Make sure that comparisons are 
   * done with .equals and not with '==' because if executions are 
   * loaded from persistent storage, a new string is created instead 
   * of the constants. */
  final String STATE_INACTIVE = "inactive";
  
  /** this execution has ended normally. Make sure that comparisons are 
   * done with .equals and not with '==' because if executions are 
   * loaded from persistent storage, a new string is created instead 
   * of the constants. */
  final String STATE_ENDED = "ended";
  
  /** this execution was cancelled with the {@link #cancel()}
   * method before normal execution ended. Make sure that comparisons are 
   * done with .equals and not with '==' because if executions are 
   * loaded from persistent storage, a new string is created instead 
   * of the constants. */
  final String STATE_CANCELLED = "cancelled";

  /** indicates that this execution is temporary suspended with the
   * {@link #suspend()} method.  Human tasks of a suspended execution
   * shouldn't show up in people's task list and timers of suspended
   * executions shouldn't fire and the execution is locked. Make sure that comparisons are 
   * done with .equals and not with '==' because if executions are 
   * loaded from persistent storage, a new string is created instead 
   * of the constants. */
  final String STATE_SUSPENDED = "suspended";

  /** the meaningless database primary key */
  long getDbid();
  
  /** the meaningful but optional business key that is unique within one 
   * process definition.  This could be for instance the order number.  
   * It's a user defined identifier for one execution within the scope of 
   * a single process definition. */
  String getKey();
  
  /** the name of this execution. The name of a main 
   * path of execution is null. */
  String getName();

  /** the <a href="#state">state</a> of this execution.  */
  String getState();

  /** is this execution active ? This is the inverse of {@link #isLocked()}. */
  boolean isActive();

  /** is this execution {@link #lock(String) locked} ? This is the inverse of {@link #isActive()}. */
  boolean isLocked();

  /** is this execution ended or cancelled ? */
  boolean isFinished();
  
  /** is this execution ended normally ? */
  boolean isEnded();

  /** is this execution suspended ? */
  boolean isSuspended();

  /** makes this execution read-only.
   * All methods that change the state of this execution (incl the fire methods) will throw 
   * a PvmException.
   * The only mutable method allowed is {@link #unlock()}.
   * @throws PvmException when this execution is already locked. */
  void lock(String state);
  
  /** unlocks a {@link #lock(String) locked} execution. */
  void unlock();
  
  /** indicates low priorities with negative values and high priorities 
   * with positive values.  The default priority is 0, which means 
   * NORMAL. Other recognized named priorities are HIGHEST (2), HIGH (1), 
   * LOW (-1) and LOWEST (-2). For the rest, the user can set any other 
   * priority integer value, but then, the UI will have to display it as 
   * an integer and not the named value.*/
  int getPriority();
  
  /** setter for the priority.  The default priority is 0, which means 
   * NORMAL. Other recognized named priorities are HIGHEST (2), HIGH (1), 
   * LOW (-1) and LOWEST (-2). For the rest, the user can set any other 
   * priority integer value, but then, the UI will have to display it as 
   * an integer and not the named value.*/
  void setPriority(int priority);

  /** the main path of execution in the <a href="package-summary.html#basicexecutionstructure">execution 
   * structure</a>.  Null will be returned in case this execution itself is the 
   * main execution path.  */
  Execution getProcessInstance();

  /** the parent execution in the <a href="package-summary.html#basicexecutionstructure">execution 
   * structure</a>.  Null will be returned in case this execution itself is the 
   * main execution path. */ 
  Execution getParent();

  /** the child executions in the <a href="package-summary.html#basicexecutionstructure">execution 
   * structure</a>.  Can be null and can be an empty collection. */ 
  Collection<Execution> getExecutions();
  
  /** maps child execution names to execution objects.  In case multiple executions 
   * have the same name, the first one is taken.  Can be null or can be an empty 
   * map.  The first execution without a name is also included with null as the key.
   */ 
  Map<String, Execution> getExecutionsMap();
  
  /** the child execution for the given name or null in case such execution doesn't exist. */ 
  Execution getExecution(String name);
  
  /** the process definition for this execution.*/
  ProcessDefinition getProcessDefinition();
  
  // current position of the execution ////////////////////////////////////////

  /** the current node indicating the position in the process definition graph.  
   * Can be null in case a transition is being taken. */ 
  Node getNode();
  
  /** the current transition indicating the position in the process definition graph.
   * Can be null in case this execution is not taking a transition. */
  Transition getTransition();
  
  /** the original object that fired the event, part of the current position in the 
   * process graph.  Can be null in case no event is being fired. This is mostly the 
   * object that is listened to with an {@link Activity}, but the eventSource can also 
   * be a child of the object to which is listened in case of event propagation. */
  ObservableElement getEventSource();
  
  /** the event that is being fired, part of the current position in the process 
   * graph.  Can be null in case no event is being fired. */
  Event getEvent();
  
  /** returns the previously executed node only if {@link Node#isPreviousNeeded()}
   * is set to true. */  
  Node getPreviousNode();
  
  /** returns the previously taken transition only if {@link Node#isPreviousNeeded()}
   * is set to true. */  
  Transition getPreviousTransition();
  
  // variables ////////////////////////////////////////////////////////////////

  /** sets the variable into the most inner matching variable scope.  If no variable 
   * scope is present, a new one is created on the process instance.  So by default, 
   * variables are created in the most outer scope and hence they will be visible 
   * for the full process instance.  For more about 
   * variables and variable scopes, see <a href="package-summary.html#variables">section 
   * variables</a> in the package description.
   * @see #removeVariable(String) 
   * @see #getVariableKeys() */
  void setVariable(String key, Object value);
  
  /** sets all the variable into the most inner matching variable scope.  This doesn't 
   * replace the variables map, so the current variables for which there is no key in 
   * the given variables map are left untouched.  Existing variables for which there is 
   * a key are overwritten.  This method just delegates to 
   * {@link #setVariable(String, Object)} for each entry in the given map.  For more 
   * about variables and their scopes, 
   * see <a href="package-summary.html#variables">section 
   * variables</a> in the package description.
   * @see #setVariable(String,Object) 
   * @see #removeVariable(String) 
   * @see #getVariableKeys() */
  void setVariables(Map<String, Object> variables);
  
  /** retrieves the variable value for the given key.  If no such key is present null 
   * will be returned.  Variables can exist and have null as value.  Variable lookup 
   * will be done according to the variable scoping rules as explained in 
   * <a href="package-summary.html#variables">section variables</a> in the package 
   * description. */
  Object getVariable(String key);
  
  /** checks presence of the given variable key.  All variable scopes are searched. */
  boolean hasVariable(String key);
  
  /** removes the variable with the given key.  Lookup of the variable will be done 
   * just like in the {@link #getVariable(String)} method.  More can be found in
   * <a href="package-summary.html#variables">section variables</a> in the package 
   * description. 
   * @see #setVariable(String, Object) */
  void removeVariable(String key);

  /** the variable keys for this execution in all visible scopes.
   * This method never returns null.  If no variables are present, an 
   * empty set will be returned.
   * @see #setVariable(String, Object) 
   * @see #removeVariable(String) */ 
  Set<String> getVariableKeys();
  
  /** retrieves all variables, CAUTION : this might be a very costly operation in 
   * case you're using database persistence and many variables have to be fetched from 
   * the database. This method never returns null.  If no variables are present, an 
   * empty map will be returned.  Changes to the returned map will not lead to updates 
   * of the variables in this execution. */
  Map<String, Object> getVariables();
  
  /** feeds a external trigger into this execution.
   * 
   * <p>Typically a signal causes the execution to proceed, but that doesn't necessarily 
   * has to be the case .  The {@link ExternalActivity} is responsible for interpreting 
   * the signal and acting upon it.
   * </p>
   * 
   * <p>A signal can optionally be given {@link #signal(String) a signal name}, 
   * {@link #signal(Map) a map of parameters} or {@link #signal(String, Map) both}.
   * </p>
   * 
   * <p>Since it's an external trigger, this method requires that this execution is 
   * waiting for an external trigger.  So this method must be called as an external client
   * and can not be called while this execution is executing.  In an {@link Activity} for 
   * example you're not allowed to call the signal on the execution cause it is executing.  
   * But you are allowed to invoke this method on any other execution (at least, if that 
   * one is waiting for an external trigger).</p>
   * 
   * <p>Typically a signal will cause the execution to start executing, but that is 
   * not a must.  What happens with this signal is defined in the 
   * {@link ExternalActivity#signal(Execution, String, Map)} of 
   * the {@link #getNode() current node}. </p>
   *  
   * @see #signal(String) */
  void signal();
  
  /** feeds a named {@link #signal() external trigger} into the execution.
   * 
   * <p>In each state, a number of things can happen.  The signal parameter specifies 
   * which of these things is happening.  It's somewhat similar to a method name in 
   * the invocation of an object.
   * </p>
   * 
   * @see #signal() See the unnamed signal for more information
   */
  void signal(String signalName);

  /** feeds {@link #signal() an external trigger} into the execution with parameters.
   * 
   * @see #signal() See the unnamed signal for more information
   */
  void signal(Map<String, Object> parameters);

  /** feeds a named {@link #signal() external trigger} into the execution with parameters.
   *
   * <p>In each state, a number of things can happen.  The signal parameter specifies 
   * which of these things is happening.  It's somewhat similar to a method name in 
   * the invocation of an object.
   * </p>
   * 
   * <p>The parameters parameter provide extra information to the signal.
   * Typically, the parameters are set as variables but 
   * the process language can overwrite that behaviour in the current node.  
   * See {@link ExternalActivity#signal(Execution, String, Map)} for more information. 
   * </p>
   * 
   * @see #signal() See the unnamed signal for more information
   */
  void signal(String signalName, Map<String, Object> parameters);

  /** ends this execution and all of its child executions.
   * 
   * <p>The execution will be removed from it's parent.  Potentially this can cause 
   * a parent execution to start executing in case this is the last concurrent 
   * execution for which the parent is waiting.</p>
   * 
   * <p>This method should not be called in {@link Activity}s.  It can be called from 
   * outside the process execution and in {@link ExternalActivity}s. </p> */
  void end();

  /** ends this execution and all it's child executions with a user defined 
   * status.  
   * 
   * <p>It is not recommended to use any of 
   * {@link #STATE_ACTIVE the defined statuses} as that may case unpredictable 
   * side effects.</p>
   *  
   * <p>The execution will be removed from it's parent.</p> */
  void end(String state);
  
  /** ends this execution and all it's child executions with a user defined 
   * status.
   * 
   * <p>It is not recommended to use any of 
   * {@link #STATE_ACTIVE the defined statuses} as that may case unpredictable 
   * side effects.</p>
   *  
   * @param remove indicates whether the execution must be removed from its parent.
   */
  void end(String state, boolean remove);
  
  /** ends this execution and assigns the state {@link #STATE_CANCELLED}.
   * @see #end(String) */
  void cancel();
  
  /** suspends this execution and all it's child executions.  Human tasks 
   * of a suspended execution shouldn't show up in people's task list and 
   * timers of suspended executions shouldn't fire. 
   * @throws PvmException if this execution is already suspended. */   
  void suspend();

  /** resumes an execution.  Inverse of {@link #suspend()}. 
   * @throws PvmException if this execution is not suspended. */ 
  void resume();
  
  // variable scopes //////////////////////////////////////////////////////////

  /** creates a new variable scope.  Variable scopes are typically managed by 
   * the process languages and users should not have to use this method. */
  VariableScope createVariableScope();

  /** creates a new variable scope and initialize it according to the given 
   * variable definitions. Variable scopes are typically managed by 
   * the process languages and users should not have to use this method. */
  VariableScope createVariableScope(List<VariableDefinition> variableDefinitions);

  /** removes the most inner variableScope. Variable scopes are typically 
   * managed by the process languages and users should not have to use this 
   * method.*/
  void destroyVariableScope();

  /** iterates over all variable scopes from inner to outer, including 
   * executionScopes on the parent execution chain. Variable scopes are 
   * typically managed by the process languages and users should not have to 
   * use this method.*/
  Iterator<VariableScope> getVariableScopeIterator();
  
  // event firing /////////////////////////////////////////////////////////////
  
  /** fires the event on the given eventSource and then propagates the event 
   * up to the eventSource's parent chain.  All the actions will see the given 
   * eventSource in {@link #getEventSource()}, event if the events are 
   * registered to parent's of the given eventSource. */
  void fire(String eventName, ObservableElement eventSource);

  // creation of child executions /////////////////////////////////////////////
 
  /** creates a child path of execution.  This execution will serve as the parent of 
   * the newly created execution.  When the child execution ends, it will be 
   * automatically removed from the list of executions.  Creating a child execution 
   * implies that this execution will no longer be a leaf in the execution tree.  
   * So the state of this execution will automatically set to {@link #STATE_INACTIVE} 
   * when this method is invoked. */
  Execution createExecution();

  /** creates a child path of execution with the given name. This execution will 
   * serve as the parent of the newly created execution.  When the created 
   * execution is ended with {@link #end()} or {@link #end(String)}, then it 
   * will be automatically removed from this parent execution. Creating a child execution 
   * implies that this execution will no longer be a leaf in the execution tree.  
   * So the state of this execution will automatically set to {@link #STATE_INACTIVE} 
   * when this method is invoked. */
  Execution createExecution(String name);

  /** removes the child execution.  Even in case the removed execution was the last child 
   * execution, the state of this execution will not be changed.  It is up to the 
   * node implementations to set the state appropriately. */
  void removeExecution(Execution execution);

  /** indicates if this execution has a child execution with the given executionName */
  boolean hasExecution(String executionName);

  // logs /////////////////////////////////////////////////////////////////////

  /** adds a <a href="package-summary.html#logs">log</a> to this execution. */
  void addLog(ProcessLog processLog);

  // extensions //////////////////////////////////////////////////////////////

  /** way to access process language extensions in the execution without 
   * having to cast.  Casting can be problematic for persistence. */
  <T> T getExtension(Class<T> extensionClass);
  
  // execution propagation methods ////////////////////////////////////////////
  
  /** takes the default transition.  
   * 
   * <p>This method can only be called from inside
   * {@link ExternalActivity} implementations and in rare occasions also from outside 
   * of the execution (from an external client while the process is in a wait state).
   * For external clients, it is more normal to use the {@link #signal()} 
   * method as in that case, it's the current node (hence the process language)that 
   * will decide how to interpret the signal.
   * </p>
   * 
   * @throws PvmException in case there is no default transition in the current node 
   * or in case this method is called from inside an {@link Activity} */
  void takeDefaultTransition();

  /** takes the outgoing transition with the given name. 
   * 
   * <p>This method can only be called 
   * from inside {@link ExternalActivity} implementations and in rare occasions also from 
   * outside of the execution (from an external client while the process is in a wait state).
   * For external clients, it is more normal to use the {@link #signal(String)} 
   * method as in that case, it's the current node (hence the process language)that 
   * will decide how to interpret the signal.</p>
   * 
   * <p>Transitions will be looked up recursively starting from the 
   * {@link #getNode() current node} and then up the {@link Node#getParent() node-parent-hierarchy} </p>
   * 
   * @param transitionName is the name of the transition to take.  A null value will 
   * match the first unnamed transition.
   *   
   * @throws PvmException in case no such transition is found in {@link #getNode() the current node} 
   * or in case this method is called from inside an {@link Activity}.*/
  void take(String transitionName);

  /** takes the given outgoing transition.  
   * 
   * <p>This method can only be called 
   * from inside {@link ExternalActivity} implementations and in rare occasions also from 
   * outside of the execution (from an external client while the process is in a wait state).
   * For external clients, it is more normal to use the {@link #signal(String)} 
   * method as in that case, it's the current node (hence the process language)that 
   * will decide how to interpret the signal.</p>
   * 
   * <p>CAUTION: It's up to the client to make sure 
   * that this transition makes sense as there is no check whether the given transition 
   * is an outgoing transition of the current node.  The motivation for that is 
   * that in case of superstates, that check can become too 'expensive'. </p> */
  void take(Transition transition);

  /** executes the given nested node. 
   * 
   * <p>The nodeName is looked up in the current node's nested nodes.</p>
   * 
   * <p>This method can only be called 
   * from inside {@link ExternalActivity} implementations and in rare occasions also from 
   * outside of the execution (from an external client while the process is in a wait state).
   * For external clients, it is more normal to use the {@link #signal(String)} 
   * method as in that case, it's the current node (hence the process language)that 
   * will decide how to interpret the signal.</p> */
  void execute(String nodeName);
  
  /** executes the given node. 
   * 
   * <p>This method can only be called 
   * from inside {@link ExternalActivity} implementations and in rare occasions also from 
   * outside of the execution (from an external client while the process is in a wait state).
   * For external clients, it is more normal to use the {@link #signal(String)} 
   * method as in that case, it's the current node (hence the process language)that 
   * will decide how to interpret the signal.</p>
   */
  void execute(Node node);
  
  /** makes this execution wait in the current node until an external trigger is given 
   * with one of the {@link #signal()} methods. */
  void waitForSignal();
  
  /** reposition this execution in the destination node. */
  void move(Node destination);
  
  /** adding a comment.  The actor making this comment is taken from the 
   * environment if it is available with {@link Environment#getUserId()}. */
  Comment createComment(String message);

  /** the list of comments made on this execution. */
  List<Comment> getComments();
  
  /** the exception in case an exception handler is handling an exception. */
  Exception getException();
}
