/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.tx;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import org.jbpm.log.Log;

import org.jbpm.env.Environment;
import org.jbpm.env.impl.DefaultEnvironment;
import org.jbpm.util.DefaultObservable;
import org.jbpm.util.Listener;
import org.jbpm.wire.WireContext;

/** simple 2 phase commit transaction.
 * no logging or recovery.
 * non thread safe (which is ok).
 * @author Tom Baeyens
 */
public class StandardTransaction extends DefaultObservable implements Transaction, Serializable, Listener {

  private static final long serialVersionUID = 1L; 
  private static Log log = Log.getLog(StandardTransaction.class.getName());
  
  public static final String STATE_ACTIVE = "active";
  public static final String STATE_COMMITTING = "committing";
  public static final String STATE_COMMITTED = "committed";
  public static final String STATE_MARKED_ROLLBACK = "marked-rollback";
  public static final String STATE_ROLLING_BACK = "rolling-back";
  public static final String STATE_ROLLEDBACK = "rolled-back";
  
  // events ///////////////////////////////////////////////////////////////////
  
  protected Environment environment;
  protected List<Resource> resources;
  protected String state = STATE_ACTIVE;
  
  public StandardTransaction() {
    log.debug("starting "+this);
  }

  public void setRollbackOnly() {
    if (state==STATE_COMMITTED) {
      throw new TransactionException("transaction was already committed");
    }
    if (state==STATE_ACTIVE) {
      state = STATE_MARKED_ROLLBACK;
    }
  }

  public boolean isRollbackOnly() {
    return ( (state==STATE_MARKED_ROLLBACK)
             || (state==STATE_ROLLEDBACK)
           );
  }

  public boolean isCommitted() {
    return (state==STATE_COMMITTED);
  }

  public boolean isRolledBack() {
    return (state==STATE_ROLLEDBACK);
  }

  public void enlistResource(Resource resource) {
    if (resources==null) {
      resources = new ArrayList<Resource>();
    }
    log.trace("enlisting resource "+resource+" to standard transaction");
    resources.add(resource);
  }

  /** This transaction will always listen to the close of the block context.
   * If the transaction is still active at that time, it will be committed.
   * If it is marked for rollback it will be rolled back.
   */
  public void event(Object source, String eventName, Object info) {
    if (WireContext.EVENT_CLOSE.equals(eventName)) {
      if (state==STATE_ACTIVE) {
        commit();
      } else {
        rollback();
      }

    } else if (DefaultEnvironment.EVENT_EXCEPTION.equals(eventName)) {
      setRollbackOnly();
    }
  }

  /** implements simplest two phase commit. */
  public void commit() {
    log.trace("committing");
    event(this, EVENT_BEFORECOMPLETION, STATE_COMMITTING);

    if (resources!=null) {

      try {
        // prepare //////////////////////////////////////////////////////////////

        // the prepare loop will be skipped at the first exception
        for (Resource resource: resources) {
          log.trace("preparing resource "+resource);
          resource.prepare();
        }

      // for any exception in the prepare phase, we'll rollback
      } catch (Throwable exception) {
        try {
          log.trace("resource threw exception in prepare.  rolling back.");
          rollback();
        } catch (Exception rollbackException) {
          log.error("rollback failed as well", rollbackException);
        }
        environment.setException(exception);
        return;
      }

      // commit ///////////////////////////////////////////////////////////////
      
      // The commit loop will try to send the commit to every resource, 
      // No matter what it takes.  If exceptions come out of resource.commit's 
      // they will be suppressed and the first exception will be rethrown after
      // all the resources are commited 
      Throwable commitException = null;
      for (Resource resource: resources) {
        try {
          log.trace("committing resource "+resource);
          resource.commit();
          
        // Exceptions in the commit phase will not lead to rollback, since some resources
        // might have committed and can't go back.
        } catch (Throwable t) {
          // TODO this should go to a special log for sys admin recovery
          log.error("commit failed for resource "+resource, t);
          if (commitException==null) {
            commitException = t;
          }
        }
      }

      if (commitException!=null) {
        environment.setException(commitException);
      }
    }
    
    state = STATE_COMMITTED;
    log.debug("committed "+this);
    fire(EVENT_AFTERCOMPLETION, STATE_COMMITTED);
  }

  public void rollback() {
    rollback(true);
  }

  public void rollback(boolean fireBeforeCompletionEvent) {
    log.trace("rolling back");
    if (fireBeforeCompletionEvent) {
      event(this, EVENT_BEFORECOMPLETION, STATE_ROLLING_BACK);
    }
    
    if (state==STATE_COMMITTED) {
      throw new TransactionException("transaction is already committed");
    }
    
    if (state!=STATE_ROLLEDBACK) {
      state = STATE_MARKED_ROLLBACK;
      
      if (resources!=null) {
        for (Resource resource: resources) {
          try {
            log.trace("rolling back resource "+resource);
            resource.rollback();
          } catch (Exception e) {
            log.error("rollback failed for resource "+resource);
          }
        }
      }

      state = STATE_ROLLEDBACK;
      log.debug("rolled back");
      fire(EVENT_AFTERCOMPLETION, STATE_ROLLEDBACK);

    } else {
      log.trace("transaction is already rolled back.  ignoring rollback.");
    }
  }
  
  List<Resource> getResources() {
    return resources;
  }

  String getState() {
    return state;
  }

  public void setEnvironment(Environment environment) {
    this.environment = environment;
  }

  public String toString() {
    return "StandardTransaction["+System.identityHashCode(this)+"]";
  }
}
