/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.wire.binding;

import org.jbpm.wire.Descriptor;
import org.jbpm.wire.descriptor.ObjectDescriptor;
import org.jbpm.wire.operation.FieldOperation;
import org.jbpm.wire.xml.WireParser;
import org.jbpm.xml.Binding;
import org.jbpm.xml.Parse;
import org.jbpm.xml.Parser;
import org.jbpm.xml.XmlUtil;
import org.w3c.dom.Element;

/**
 * <p>This {@link Binding} specifies a {@link FieldOperation} (for field injection).</p>
 *
 * <p>A field injection is defined by a <b><code>{@literal <field>}</code></b> xml element.</p>
 *
 * <p>This element has a attribute "name", which specifies the name of the field to inject.</p>
 * <p>This element should contain a child element that defines a {@link Descriptor}.
 * This descriptor will be used to create the object that will be assigned to the field</p>
 *
 * <h3>Example</h3>
 *
 * Consider the following class:
 * <pre> public class Foo {
 *   String bar;
 * }</pre>
 *
 * The following Xml declaration will create an object 'o' of class 'Foo' (see {@link ObjectDescriptor}), and
 * the value <code>hello</code> will be assigned to <code>o.bar</code>.
 *
 * <pre> &lt;objects&gt;
 *   &lt;object name='o' class='Foo'&gt;
 *     &lt;field name='bar'&gt;
 *       &lt;string value='hello' /&gt;
 *     &lt;/field&gt;
 *   &lt;/object&gt;
 * &lt;/objects&gt;</pre>
 *
 * @author Tom Baeyens
 * @author Guillaume Porcher (documentation)
 *
 * @see ObjectBinding
 * @see WireParser
 */
public class FieldBinding implements Binding {

  public Object parse(Element element, Parse parse, Parser parser) {
    FieldOperation fieldOperation = new FieldOperation();
    if (element.hasAttribute("name")) {
      fieldOperation.setFieldName(element.getAttribute("name"));
    } else {
      parse.addProblem("field must have name : "+XmlUtil.toString(element));
    }
    Element descriptorElement = XmlUtil.element(element);
    if (descriptorElement!=null) {
      Descriptor descriptor = (Descriptor) parser.parseElement(descriptorElement, parse, WireParser.CATEGORY_DESCRIPTOR);
      if (descriptor!=null) {
        fieldOperation.setDescriptor(descriptor);
      } else {
        parse.addProblem("unknown descriptor element "+descriptorElement.getTagName()+" inside field operation: "+XmlUtil.toString(element));
      }
    } else {
      parse.addProblem("field must have 1 descriptor element out of "+parser.getBindings().getTagNames(WireParser.CATEGORY_DESCRIPTOR)+" as content: "+XmlUtil.toString(element));
    }
    return fieldOperation;
  }
}