/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.wire.descriptor;

import java.sql.Connection;
import org.jbpm.log.Log;

import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.jbpm.env.Environment;
import org.jbpm.hibernate.HibernateTransactionResource;
import org.jbpm.tx.StandardTransaction;
import org.jbpm.util.Listener;
import org.jbpm.wire.WireContext;
import org.jbpm.wire.WireDefinition;
import org.jbpm.wire.WireException;

/**
 * @author Tom Baeyens
 */
public class HibernateSessionDescriptor extends AbstractDescriptor {
  
  private static final long serialVersionUID = 1L;
  private static final Log log = Log.getLog(HibernateSessionDescriptor.class.getName());
  
  protected String factoryName;
  protected boolean useCurrent = false;
  protected boolean tx = true;
  protected boolean close = true;
  protected String standardTransactionName;
  protected String connectionName;

  public Object construct(WireContext wireContext) {
    // get the hibernate-session-factory
    SessionFactory sessionFactory = null;
    if (factoryName!=null) {
      sessionFactory = (SessionFactory) wireContext.get(factoryName);
    } else {
      Environment environment = wireContext.getEnvironment();
      if (environment!=null) {
        sessionFactory = environment.get(SessionFactory.class);
      } else {
        sessionFactory = wireContext.get(SessionFactory.class);
      }
    }
    if (sessionFactory==null) {
      throw new WireException("couldn't find hibernate-session-factory "+(factoryName!=null ? "'"+factoryName+"'" : "by type ")+"to open a hibernate-session");
    }

    // open the hibernate-session
    Session session = null;
    if (useCurrent) {
      log.debug("getting current hibernate session");
      session = sessionFactory.getCurrentSession();
      
    } else if (connectionName!=null) {
      Connection connection = (Connection) wireContext.get(connectionName);
      log.debug("creating hibernate session with connection "+connection);
      session = sessionFactory.openSession(connection);

    } else {
      log.debug("creating hibernate session");
      session = sessionFactory.openSession();
    }

    return session;
  }

  public void initialize(Object object, WireContext wireContext) {
    Session session = (Session) object;
    
    // if transactions are enabled
    if (tx) {
      log.debug("beginning hibernate transaction");
      org.hibernate.Transaction hibernateTransaction = session.beginTransaction();
      
      // get the standard-transaction
      StandardTransaction standardTransaction = null;
      if (standardTransactionName!=null) {
        standardTransaction = (StandardTransaction) wireContext.get(standardTransactionName);
      } else {
        standardTransaction = wireContext.get(StandardTransaction.class);
      }
      if (standardTransaction==null) {
        throw new WireException("couldn't find standard-transaction "+(standardTransactionName!=null ? "'"+standardTransactionName+"'" : "by type ")+"to enlist the hibernate transaction");
      }

      // enlist the hibernate transaction to the global transaction
      HibernateTransactionResource resource = new HibernateTransactionResource(hibernateTransaction, session);
      standardTransaction.enlistResource(resource);
    }

    if (close) {
      // make sure that the session is closed when the context closes.
      // method event will be called (see below)
      wireContext.addListener(new SessionCloser(session));
    }
  }
  
  public Class<?> getType(WireDefinition wireDefinition) {
    return Session.class;
  }

  public static class SessionCloser implements Listener {
    Session session;
    public SessionCloser(Session session) {
      this.session = session;
    }
    public void event(Object source, String eventName, Object info) {
      if (WireContext.EVENT_CLOSE.equals(eventName)) {
        log.trace("closing hibernate session");
        session.close();
      }
    }
  }

  public void setFactoryName(String factoryName) {
    this.factoryName = factoryName;
  }
  public void setTx(boolean tx) {
    this.tx = tx;
  }
  public void setStandardTransactionName(String standardTransactionName) {
    this.standardTransactionName = standardTransactionName;
  }
  public void setConnectionName(String connectionName) {
    this.connectionName = connectionName;
  }
  public void setUseCurrent(boolean useCurrent) {
    this.useCurrent = useCurrent;
  }
  public void setClose(boolean close) {
    this.close = close;
  }
}
