/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.xml;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import javax.xml.namespace.QName;
import org.w3c.dom.Element;

/** a set of {@link Binding}s divided into categories managed by a {@link Parser}.
 * 
 * <a href="./Parser.html#bindings">See also 'Bindinds'</a>
 * @see Parser
 * @author Tom Baeyens
 */
public class Bindings {

  private static final String DEFAULT_CATEGORY = "default";
  protected Map<String, Map<QName,Binding>> categories = null;
  
  public Bindings() {
  }

  /** to be used when you want to customize the binding behaviour of a {@link Parser}. */
  public Bindings(Bindings other) {
    if (other.categories!=null) {
      categories = new HashMap<String, Map<QName,Binding>>();
      Set<String> categorieNames = other.categories.keySet();
      if (categorieNames!=null) {
        for (String categoryName: categorieNames) {
          Map<QName,Binding> categoryBindings = other.categories.get(categoryName);
          HashMap<QName, Binding> categoryBindingsCopy = new HashMap<QName, Binding>(categoryBindings);
          categories.put(categoryName, categoryBindingsCopy);
        }
      }
    }
  }

  /** get a binding for the given element and category. 
   * If the category is null, then all the categories will be searched for a binding in random order.
   */
  public Binding getBinding(Element element) {
    return getBinding(element, null);
  }

  /** get a binding for the given element and category. 
   * If the category is null, then all the categories will be searched for a binding in random order.
   */
  public Binding getBinding(Element element, String category) {
    String tagName = XmlUtil.getTagName(element);
    QName elementQName = new QName(element.getNamespaceURI(),tagName);
    return getBinding(elementQName, category);
  }

  /** looks up a binding for a given tagName in the default category. */
  public Binding getBinding(QName tagName) {
    return getBinding(tagName, DEFAULT_CATEGORY);
  }

  /** looks up a binding for a given tagName and category. */
  public Binding getBinding(QName tagName, String category) {
    if (categories!=null) {
      // if no category is specified
      if (category==null) {
        // search in all categories
        for (Map<QName, Binding> categoryBindings : categories.values()) {
          Binding binding = categoryBindings.get(tagName);
          if (binding != null) {
            return binding;
          }
        }
      }
    }
    Map<QName, Binding> categoryBindings = (categories!=null ? categories.get(category) : null);
    return (categoryBindings!=null ? categoryBindings.get(tagName) : null);
  }

  /** add an elementParser to this parser that will handle parsing of 
   * elements of the given tagName for the default category.
   */ 
  public void addBinding(String tagName, Binding binding) {
    addBinding(new QName(tagName), binding, DEFAULT_CATEGORY);
  }

  public void addBinding(QName tagName, Binding binding) {
    addBinding(tagName, binding, DEFAULT_CATEGORY);
  }

  /** add an elementParser to this parser that will handle parsing of 
   * elements of the given tagName for the given category.
   */ 
  public void addBinding(String tagName, Binding binding, String category) {
    addBinding(new QName(tagName), binding, category);
  }
  
  public void addBinding(QName tagName, Binding binding, String category) {
    if (categories==null) {
      categories = new HashMap<String, Map<QName,Binding>>();
    }
    Map<QName, Binding> categoryBindings = categories.get(category);
    if (categoryBindings==null){
      categoryBindings = new HashMap<QName,Binding>();
      categories.put(category, categoryBindings);
    }
    categoryBindings.put(tagName, binding);
  }
  
  /** the set of all tagNames for which there is a binding specified in the given category */
  public Set<QName> getTagNames(String category) {
    Map<QName, Binding> categoryBindings = (categories!=null ? categories.get(category) : null );
    Set<QName> bindingTags = (categoryBindings!=null ? categoryBindings.keySet() : null);
    return bindingTags;
  }
}
